/*
 * Copyright (c) Developers of JTowers
 * Redistributable and modifiable. Please see license.readme for details.
 *
 * ammendment history
 * <date>			<changed by>		<summary of changes>
 * 18 August 2004	Jonee Ryan Ty		First documented history
 *
 */

import javax.swing.JSlider;
import javax.swing.JOptionPane;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JCheckBox;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;
import javax.swing.JOptionPane;
import javax.swing.SwingConstants;

import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Cursor;
import java.awt.FlowLayout;
import java.awt.Container;

import java.net.URL;


public class App		extends javax.swing.JApplet
						implements 	javax.swing.event.ChangeListener,
									java.awt.event.ActionListener
{
	public JLabel timerLabel, numOfMovesLabel, timerHeader;
	public Image pics[];
	
	public JButton undo, redo;
	
	private ButtonGroup grp;
	private JPanel p1, p2, p3, p4;
	        
	private JButton restart, solve, step, back;
		
	private JSlider numOfDisks, speed;
	private JCheckBox constrained, randomized, spread;

	private HanoiCanv hanoiCanv;

	private boolean isIterating = false;
	private int iteration = 0;
	
	public App() {		
		super();
		
		this.setSize(Constants.APP_WIDTH, Constants.APP_HEIGHT);
		this.setCursor(new Cursor(Cursor.HAND_CURSOR));
	}

	public void init() {
		super.init();

		URL thisUrl = this.getDocumentBase();

		pics = new Image[3];
				// may take time to load
		pics[0] = this.getImage(thisUrl, Constants.BACK_PIX);
		pics[1] = this.getImage(thisUrl, Constants.DISK_PIX);
		pics[2] = this.getImage(thisUrl, Constants.DISK_PIX2);

		MediaTracker tracker = new MediaTracker(this);
		tracker.addImage(pics[0], 0);
		tracker.addImage(pics[1], 1);
		tracker.addImage(pics[2], 2);

      	try {
      		tracker.waitForAll();
      	
      	} catch(InterruptedException ie) {}

		this.setup();
	}
		// threads should be destroyed
	public void destroy() {
		hanoiCanv.destroy();
	}

		// do something with timerThread when focus leaves applet window
	public void stop() {	
		if (!hanoiCanv.isSolving)	//leave alone when solving
			hanoiCanv.stop();
	}		
		// update timerThread	
	public void start() {
		if (!hanoiCanv.isSolving)	//leave alone when solving
			hanoiCanv.start();
	}
	
	public void setup() {
		grp = new ButtonGroup();

   		constrained = new JCheckBox("constrained", false);
   		constrained.setMnemonic('c');
   		constrained.setToolTipText(
   			"source to destination, alternate to source not allowed");
   		constrained.addActionListener(this);
   		
   		randomized = new JCheckBox("randomized", false);
   		randomized.setMnemonic('r');
   		randomized.setToolTipText("play the randomized version");
   		randomized.addActionListener(this);   		
   		
   		spread = new JCheckBox("spread", false);
   		spread.setMnemonic('s');
   		spread.setToolTipText("spread the disks into the three pegs");
   		spread.addActionListener(this);   		  

		p1 = new JPanel(new FlowLayout(FlowLayout.CENTER, 40, 0));
		p1.add(constrained);	
		p1.add(randomized);					
		p1.add(spread);			
		
		numOfDisks = new JSlider(JSlider.HORIZONTAL, 2, 15, 4);
		numOfDisks.setMajorTickSpacing(3);
		numOfDisks.setMinorTickSpacing(1);
		numOfDisks.setPaintTicks(true);
		numOfDisks.setPaintLabels(true);
		numOfDisks.addChangeListener(this);

		speed = new JSlider(JSlider.HORIZONTAL, 0, 20, 10);
		speed.setMajorTickSpacing(4);
		speed.setMinorTickSpacing(1);
		speed.setPaintTicks(true);
		speed.setPaintLabels(true);
		speed.addChangeListener(this);

		p2 = new JPanel(new FlowLayout(FlowLayout.CENTER, 10, 0));
		p2.add(new JLabel("Number of disks"));
		p2.add(numOfDisks);
		p2.add(new JLabel("Delay per move (in 0.1 s)"));
		p2.add(speed);

		restart = new JButton("restart");
		restart.setToolTipText("Redraw the canvas.");
		restart.addActionListener(this);
		restart.setMnemonic('s');
		
		step = new JButton("step");
		step.setToolTipText("See the solution a step at a time");
		step.addActionListener(this);
		step.setMnemonic('t');		

		back = new JButton("back");
		back.setToolTipText("Undo last step");
		back.addActionListener(this);
		back.setMnemonic('b');
		
		solve = new JButton("solve");
		solve.setToolTipText("See the solution.");
		solve.addActionListener(this);
		solve.setMnemonic('v');

		redo = new JButton("redo");
		redo.setToolTipText("Redo move");
		redo.addActionListener(this);
		redo.setMnemonic('e');

		undo = new JButton("undo");
		undo.setToolTipText("Undo last move");
		undo.addActionListener(this);
		undo.setMnemonic('u');

		p3 = new JPanel(new FlowLayout(FlowLayout.CENTER, 14, 0));
		p3.add(restart);
		p3.add(solve);
		p3.add(redo);
		p3.add(undo);
		p3.add(step);
		p3.add(back);

		timerLabel = new JLabel("0", SwingConstants.RIGHT);
		numOfMovesLabel = new JLabel("0", SwingConstants.RIGHT);

		p4 = new JPanel(new FlowLayout(FlowLayout.CENTER, 20, 0));
		timerHeader = new JLabel("Timer (in 1 s):");
		p4.add(timerHeader);
		p4.add(timerLabel);
		p4.add(new JLabel("Number of moves:"));
		p4.add(numOfMovesLabel);

		hanoiCanv = new HanoiCanv(this);
		
		hanoiCanv.isConstrained = constrained.isSelected();
		hanoiCanv.isRandomized = randomized.isSelected();
		hanoiCanv.isSpread = spread.isSelected();		
		
		this.newGame();

		Container contentPane = this.getContentPane();
		contentPane.setLayout(
			new FlowLayout(FlowLayout.CENTER, this.getWidth(), 8));

		contentPane.add(p1);
		contentPane.add(p2);
		contentPane.add(p3);
		contentPane.add(p4);
		contentPane.add(hanoiCanv);
				
		hanoiCanv.showMessage(Constants.COPYRIGHT);	
		speed.setToolTipText(hanoiCanv.delay + " milliseconds");
	} // end of setup()

	public void actionPerformed(java.awt.event.ActionEvent e) {
		Object source = e.getSource();
		
		handleMsg();
		
		if(source instanceof JButton) {
		
			if (source == step) 
				handleIteration();				
				
			else if (source == back)
				handleBack();
					
			else
				stopIteration();
			
			if(source == restart) {
				timerHeader.setText("Timer (in 1 s):");
				this.newGame();			
				
			} else if(source == solve) {
				
				if (hanoiCanv.delay != Constants.FASTEST)
					timerHeader.setText("Timer (in 0.1 s):");
						
				else
					timerHeader.setText("Timer (in 0.01 s):");
			
				hanoiCanv.solve(
					Math.max(speed.getValue() * 100, Constants.FASTEST), 
					numOfDisks.getValue());

			} else if (source == redo) {
				if (hanoiCanv.isSolving)
					return;
					
				hanoiCanv.doRedo();	

			} else if (source == undo) {
				if (hanoiCanv.isSolving)
					return;			

				hanoiCanv.doUndo();	
			}
		
		} else if (source instanceof JCheckBox) {
			
			stopIteration();
			
			if (source == constrained) {
				
				hanoiCanv.isConstrained = constrained.isSelected();	
				
				if (hanoiCanv.isSolving) {
					hanoiCanv.solve(
						Math.max(speed.getValue() * 100, Constants.FASTEST), 
						numOfDisks.getValue());				
				
				} else
					hanoiCanv.repaint(0, 0, HanoiCanv.BACK_WIDTH, 50);
			
			} else if (source == spread) {
				
				hanoiCanv.isSpread = spread.isSelected();
				
				if (hanoiCanv.isSolving)
					hanoiCanv.solve(
						Math.max(speed.getValue() * 100, Constants.FASTEST), 
						numOfDisks.getValue());				
				else 
					newGame();
			
			} else if (source == randomized) {
				hanoiCanv.isRandomized = randomized.isSelected();
				
				if (hanoiCanv.isSolving)
					hanoiCanv.solve(
						Math.max(speed.getValue() * 100, Constants.FASTEST), 
						numOfDisks.getValue());				
				else 
					newGame();			
			}
		}
	
	} // actionPerformed

		// processes events generated by JSlider objects
	public void stateChanged(javax.swing.event.ChangeEvent e) {
		if(!((JSlider)(e.getSource())).getValueIsAdjusting())
		{
			if(e.getSource() == speed) {
				hanoiCanv.setDelay(
					Math.max(speed.getValue() * 100, Constants.FASTEST));

				if (hanoiCanv.isSolving) {
					
					if (hanoiCanv.delay != Constants.FASTEST)
						timerHeader.setText("Timer (in 0.1 s):");
						
					else
						timerHeader.setText("Timer (in 0.01 s):");
				}
					
				speed.setToolTipText(hanoiCanv.delay + " milliseconds");
			}

			else {// e.getSource() == numOfDisks
				stopIteration();		
				handleMsg();
				this.newGame();				
			}
		}
	}

	/**
	 * sets the canvas
	 */

	private void newGame() {

		hanoiCanv.newGame(Math.max(speed.getValue() * 100, Constants.FASTEST),
			numOfDisks.getValue());
                
 		undo.setEnabled(false);               
 		redo.setEnabled(false);
 		back.setEnabled(false);
	}
	
	private void stopIteration() {
		if (isIterating) {
			iteration = 0;
			isIterating = false;
			
			hanoiCanv.showMessage("iteration stopped");
		}
	}
	
	private void handleIteration() {
		if (!isIterating) {
			isIterating = true;
			hanoiCanv.firstStep(
				Math.max(speed.getValue() * 100, Constants.FASTEST),
				numOfDisks.getValue());											
		} else
			hanoiCanv.iteration(iteration);
			
		iteration++;	
		back.setEnabled(true);
	}
	
	private void handleBack() {
		iteration--;
				
		if (iteration == 0)
			back.setEnabled(false);
			
		hanoiCanv.showMessage("back" + Integer.toString(iteration));		
	}
		
	private void handleMsg() {
		hanoiCanv.showCopyright = false;
	}
	
} //##30
