/*
 * Copyright (c) Developers of JTowers
 * Redistributable and modifiable. Please see license.readme for details.
 *
 * ammendment history
 * <date>			<changed by>		<summary of changes>
 * 18 August 2004	Jonee Ryan Ty		First documented history
 *
 */
 
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Color;
import java.awt.Graphics;

import java.awt.event.MouseEvent;

import java.util.Date;
import java.util.EmptyStackException;

import javax.swing.JLabel;
import javax.swing.JOptionPane;

/**
 *the drawing board for JTower
 */
 
public class HanoiCanv		extends javax.swing.JPanel				
							implements 	java.awt.event.MouseListener, 
										java.awt.event.MouseMotionListener
{	
	public static final int	BACK_WIDTH	= 450,	// dimensions of this pane
							BACK_HEIGHT	= 280;
								
	public static final int ORIGINAL 	= 12,
							RANDOMIZED 	= 13,
							
							SANDY		= 0,
							ANDY		= 1,
							DANDY		= 2,							
							
							LEFT_CLICK = 16,	// mouse event modifiers
							RIGHT_CLICK = 4;							
			// the following are public so that access from SolveThread ok
	public JLabel numOfMovesLabel, timerLabel, timerHeader;
	
	public int	movesMade = 0,
				numOfDisks,
				delay, 		// animation delay
				permut[],
				spread[];	// used for RANDOMIZED2 game and spread
	
	public Peg	sandy = null,	
				andy = null,
				dandy = null;
	
	public Disk selected;
	
	public boolean 	isSolving = false, 
					aDiskSelected = false,
						// block also mouse events when game is over	
					mouseEventsAreBlocked = false,
					isConstrained = false, // if playing a constrained game
					isSpread = false,
					isRandomized = false,
					showMessage = false,
					showCopyright = true;
	
	public String msg;	
						
	public volatile SolveThread solve;
	private volatile TimerThread timerThread;
	private volatile SetupThread setup;
	
	private boolean isFirstMove, // after first move, start timer
					gameIsOver = false,
					timerSuspended = false,
					drawOutline = true;
													
	private Stack undo, redo;
	private Image pics[], background, disk_pix, disk_pix2;
	private App parent;	
	private long prevTime;		
	private int xMo, yMo, xExtra, yExtra, // mouse coordinates and allowances
				action = SetupThread.SETUP;	
	
	public class UndoRedoEntry {
		
		public Peg from, to;
		public Disk disk;
		
		public UndoRedoEntry(Peg from, Peg to, Disk disk) {
			this.from = from;
			this.to = to;
			this.disk = disk;
		}
		
	}
	
	public HanoiCanv(App parent) {
		super();
		
		this.parent = parent;
		this.pics = parent.pics;
		this.background = pics[0];
		this.disk_pix = pics[1];
		this.disk_pix2 = pics[2];
		
		this.timerLabel = parent.timerLabel;
		this.timerHeader = parent.timerHeader;
		this.numOfMovesLabel = parent.numOfMovesLabel;				
        		
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.setPreferredSize(new java.awt.Dimension(BACK_WIDTH, BACK_HEIGHT));
		
		undo = StackFactory.createStack();
		redo = StackFactory.createStack();
	}

	private void drawDisk(Graphics g, Disk d, boolean isSelected) {
		
		if (d == null) 
			return;
		
		String toWrite;
		Image toDraw;
		
		if (!isSelected)
			toDraw = disk_pix;
		else
			toDraw = disk_pix2;
		
		g.drawImage(toDraw, d.x + 2, d.y, d.getWidth() - 4, Disk.HEIGHT, this);
	
		if (drawOutline) {
			//g.drawRect(d.x, d.y, d.getWidth(), Disk.HEIGHT);
			toWrite = Integer.toString(d.getDiskNum());
				
			g.drawString(toWrite, d.x + d.getWidth() / 2 - toWrite.length() * 3,
				d.y + Disk.HEIGHT / 2 + 3);			
		}
		
	}
	
	private void drawPeg(Graphics g, Peg p) {
			
		if (p == null)
			return;
		
		StackNode tempNode;
		Disk tmpDisk;
			
		if(p.size() != 0) {	//if peg is not empty

			tempNode = (StackNode)(p.getTopStackNode());
					//draw each disk
			while(tempNode != null) {
				tmpDisk = (Disk)(tempNode.getData());
				drawDisk(g, tmpDisk, false);
				tempNode = tempNode.getNext();
			}
		}						
	}
	
	public void paintComponent(Graphics g) {

		super.paintComponent(g);

		g.setColor(Color.cyan);
						// draw background image
		g.drawImage(background, 0, 0, BACK_WIDTH, BACK_HEIGHT, this);
			
						//to draw all three pegs
		drawPeg(g, sandy);
		drawPeg(g, andy);
		drawPeg(g, dandy);		

			//draw selected disk ie when dragging
		if (aDiskSelected)
			drawDisk(g, selected, true);			
			
		if (showMessage) {
			g.drawString(msg, 10, 15);
			showMessage = false;
		
		} else if (gameIsOver)
			g.drawString(Constants.CONGRATULATIONS, 10, 15);
		
		else if	(showCopyright)
			g.drawString(Constants.COPYRIGHT, 10, 15);
		

		
	} // end of paintComponent
    
    public void mouseClicked(MouseEvent e) {}
	
	public void mouseEntered(MouseEvent e) {}
   
	public void mouseExited(MouseEvent e) {}
	
		//returns true when the user clicked on the area of the parameter peg
	private boolean handleClickOnATopDisk(Peg p) {
		
		Disk topDisk;	
		boolean onTopOfPeg = false;				
					//check if user clicks on a top disk
		if (onTopOfPeg(p)) {
			
			onTopOfPeg = true;
			
			if (p.size() != 0) {
				topDisk = p.top();
				
				if (onTopOfDisk(topDisk)) {
					aDiskSelected = true;
					selected = topDisk;		
							
				// make a reference to the previous	peg that contained disk
					selected.whereContained = p;	
					p.pop(); // the popped element is the same as selected
					
					xExtra = selected.x - xMo;
					yExtra = selected.y - yMo;
				}				
			}
		}

		return onTopOfPeg;		
	}
	
	public void mousePressed(MouseEvent e) {	
		
		if((e.getModifiers() == RIGHT_CLICK) || 
				mouseEventsAreBlocked)
			return;
			
		xMo = e.getX();
     	yMo = e.getY();	
		
		if (handleClickOnATopDisk(sandy)); 
		
		else if (handleClickOnATopDisk(andy));
		
		else if (handleClickOnATopDisk(dandy));
		
	} // mousePressed()
	
	public void mouseDragged(MouseEvent e) {
		if ((e.getModifiers() == RIGHT_CLICK) || 
				mouseEventsAreBlocked)
			return;
						
		xMo = e.getX();
     	yMo = e.getY();
     	
     		//"release" mouse if it goes out of canvas while dragging a disk
		if (!onTopOfRect(0, 0, BACK_WIDTH, BACK_HEIGHT) && (aDiskSelected)) { 
			this.mouseReleased(new MouseEvent(this, 
				MouseEvent.MOUSE_RELEASED, (new  Date()).getTime(),
				MouseEvent.BUTTON1_MASK, xMo, yMo, 1, false));
			return;	
		}				
					
		if (aDiskSelected) {					

			selected.x = xMo + xExtra;
			selected.y = yMo + yExtra;
			
			repaint();
		}
	}
		// to destroy an existing TimerThread		
	private void destroyTimerThread() {		
		
		if(timerThread != null) {			
			try {				
				if (timerThread.isAlive())			
					timerThread.interrupt();
			} catch(SecurityException se) {}
			
			timerThread.timerLabel = null;
			timerThread = null;
		}	
	}
	
	private void handleGameOver() {
	
		// puzzle solved when all disks are in dandy
		if((andy.size() == 0) && (sandy.size() == 0)) {
			destroyTimerThread();
			
			mouseEventsAreBlocked = true;
			gameIsOver = true;
			
			repaint(0, 0, BACK_WIDTH, 50);	// to show congratulatory remarks
			parent.undo.setEnabled(false);
			parent.redo.setEnabled(false);
		}			
	}
		// returns true when a disk drops within the boundary of a peg
	private boolean handlePegDrop(Peg p) {

		boolean isOnTopOfPeg = false;
		
		if (onTopOfPeg(p)) {

			isOnTopOfPeg = true;
			boolean moveIsValid = true;

				// same peg as preivous
			if(selected.whereContained == p)
				moveIsValid = false;

		//additionally consider the constrains of a HanoiCanv.CONSTRAINED game
			if (isConstrained) {
				if (selected.whereContained == sandy && p == dandy) 
					moveIsValid = false;
				else if (selected.whereContained == andy && p == sandy)
					moveIsValid = false;					
			}
	
				// move is invalid when top is smaller than selected disk
			if (p.size() != 0)
				if(p.top().getDiskNum() < selected.getDiskNum())
						moveIsValid = false;
			
			if (!moveIsValid)
				HanoiCanv.pushBack(selected);
			
			else {		

				redo = StackFactory.createStack();

				UndoRedoEntry toPush = new UndoRedoEntry(
					selected.whereContained, p, selected);
				undo.push(toPush);					
				
				movesMade++;
				numOfMovesLabel.setText(Integer.toString(movesMade));
				
				HanoiCanv.push(selected, p);	// push the released disk

				if (p == dandy)
					handleGameOver();					
			}
		}
							
		return isOnTopOfPeg;	
	}

	private void setPlayTime() {		
		timerHeader.setText("Timer (in s):");
	}
	
	public void mouseReleased(MouseEvent e) {
		
		if ((e.getModifiers() == RIGHT_CLICK) || 
				mouseEventsAreBlocked)
			return;
		
		if (!aDiskSelected)
			return;
		
		if (isFirstMove) {		// start timer at first move			
			destroyTimerThread();			
			timerThread = new TimerThread(timerLabel, 0, 1000);
			timerThread.start();					
			isFirstMove = false;
			setPlayTime();			
		}	
		
		xMo = e.getX();
     	yMo = e.getY();	
		
		aDiskSelected = false;
		
		int prevX = selected.x,	// previous coordinates of disk
		 	prevY = selected.y;

		 				// move is invalid ONLY when top is smaller 
		boolean moveIsValid = true;		// than selected disk.
		
		if (handlePegDrop(dandy));
		
		else if (handlePegDrop(andy));
		
		else if (handlePegDrop(sandy));
					
		else 	//if it is not dropped to a peg boundary push it back to prev
			HanoiCanv.pushBack(selected);
							
	//draw disk at previous space then repaint new location occupied by disk
		repaint(prevX, prevY, selected.getWidth(), Disk.HEIGHT);
		repaint(selected.x, selected.y, selected.getWidth(), Disk.HEIGHT);

		enableRedoUndo();			
			
	} // end of mouseReleased
	
	private void enableRedoUndo() {
		parent.undo.setEnabled(undo.size() != 0);
		parent.redo.setEnabled(redo.size() != 0);		
	}
	
	public void mouseMoved(MouseEvent e) {}
		
		// destroy all threads
	public void destroy() {
				
		if (setup != null) {
			try {
				if (setup.isAlive()) {
					setup.interrupt();
					setup.hanoiCanv = null;
					setup = null;
				}		
			} catch (Exception ex) {
				ex.printStackTrace();
				System.out.println("caught\n");				
			}
		}
				
		if(solve != null) {			
			try {							
				if (solve.timerThread != null) {
					if (solve.timerThread.isAlive()) {
						solve.timerThread.interrupt();		
						solve.timerThread.timerLabel = null;
						solve.timerThread = null;
					}
				}
			} catch (Exception ex) {
				ex.printStackTrace();
				System.out.println("caught\n");				
			}
			
			try {
				if (solve.isAlive()) {
					solve.interrupt();			
					solve.hanoiCanv = null;
					solve = null;
				}				
			} catch (Exception ex) {
				ex.printStackTrace();			
				System.out.println("caught\n");				
			}
		}	
		
		destroyTimerThread();
		isSolving = false;
	}
	
		// do something with timerThread when focus leaves parent applet window
	public void stop() {	
		// Thread method stop() is unsafe and at j2sdk1.4 it is deprecated
		if(timerThread != null) {			
			prevTime = Long.parseLong(timerLabel.getText());			
			destroyTimerThread();		
			timerSuspended = true;
		}
	}		
		// update timerThread when applet is refocused	
	public void start() {
		
		if(timerSuspended) {
			timerThread = new TimerThread(timerLabel, prevTime);
			timerThread.start();
			
			timerSuspended = false;
		}
	}
	 	
	private void jumble(int permut[]) {
		
		if (permut.length == 1)
			return;
		
		int swpIndx, tmp;
							// jumble array to get permutation
		for(int b = 0; b < permut.length; b++) {
			swpIndx = (int)((numOfDisks) * Math.random());
			
			tmp = permut[b];
			
			permut[b] = permut[swpIndx];
			permut[swpIndx] = tmp;
		}				
	}
	
	private void setPermut() {
						
		permut = new int[numOfDisks];
		
		for (int a = 0; a < numOfDisks; a++)
			permut[a] = a + 1;
			
		if (isRandomized  && (numOfDisks != 1))
			jumble(permut);
	}
		// will allocate disks to three pegs in RANDOMIZED2 game
	private void setSpread() {
			
		int random;		
		spread = new int[numOfDisks];

		if (!isSpread) 
			for (int b = 0; b < numOfDisks; b++)
				spread[b] = HanoiCanv.SANDY;	
			
		else 
			for (int b = 0; b < numOfDisks; b++)
				spread[b] = 
					(int)(Math.round(Math.random() * Long.MAX_VALUE) % 3);
	}

	public void solve(int delay, int numOfDisks) {
		newGame(delay, numOfDisks, SetupThread.WILL_SOLVE_NEXT);
	}

	public void newGame(int delay, int numOfDisks) {
		newGame(delay, numOfDisks, SetupThread.SETUP);
	}
	
	public void firstStep(int delay, int numOfDisks) {
		newGame(delay, numOfDisks, SetupThread.WILL_STEP);
	}
	
	public void iteration(int n) {
		showMessage(Integer.toString(n));
	}
	
	public void showMessage(String msg) {
		this.msg = msg;
		this.showMessage = true;	
		repaint(0, 0, BACK_WIDTH, 50);		
	}

	/**
	 *sets up this canvas
	 */
	 	 
	private void newGame(int delay, int numOfDisks, int action) {		
		this.delay = 1000;	// buffer time so that threads may be interrupted
		destroy();	// to destroy any existing thread				
		
		this.numOfDisks = numOfDisks;
		this.delay = delay;

		this.action = action;
		setup = new SetupThread(this, action);
		setup.start();
	}
	 
	public void setup() {
	
		sandy = new Peg(0, 25, "sandy");
		andy = new Peg(150, 25, "andy");
		dandy = new Peg(300, 25, "dandy");
		selected = null;
		
		System.gc();

		movesMade = 0;
		numOfMovesLabel.setText("0");		
			
		isFirstMove = true;

		if (gameIsOver) {
			gameIsOver = false;
			repaint(0, 0, BACK_WIDTH, 50);
		}
		
		this.timerLabel.setText("0");
	
		undo = StackFactory.createStack();
		redo = StackFactory.createStack();

		if (action != SetupThread.WILL_SOLVE_NEXT) {
			setPermut();
			setSpread();
		}
		
			//repaint only the area occupied by the three pegs
		repaint(sandy.getX(), sandy.getY(), BACK_WIDTH, Peg.HEIGHT);

	} // setup
	
	public void setDelay(int newDelay) {
			// get current base
		boolean toLesserBase = (this.delay == Constants.FASTEST);	
		
		if (newDelay == Constants.FASTEST && this.delay > Constants.FASTEST)
			toLesserBase = true;	// to lesser base
		else if (this.delay == Constants.FASTEST && 
				newDelay > Constants.FASTEST)
			toLesserBase = false;	// to bigger base

		if (solve != null) {
			//change time to base 0.01 or 0.1 whatever applicable
			solve.changeTimeBase(toLesserBase);
		}
		
		this.delay = newDelay;				
	}
	
	public static void pushBack(Disk d) {
		HanoiCanv.push(d, d.whereContained);
	}
		
	public static void push(Disk d, Peg p) {
		
		d.x = p.getMidX() - d.getWidth() / 2;
		d.y = p.getBaseY() - (1 + p.size()) * Disk.HEIGHT;
	
		p.push(d);
	}
	
	public void doRedo() {
		
		if (redo.size() == 0)
			return;

		UndoRedoEntry tmp = (UndoRedoEntry)redo.pop();
		
		if (tmp.from.top() == tmp.disk) {
			Disk toRedo = tmp.from.pop();
			int prevX = toRedo.x;
			int prevY = toRedo.y;
			
			movesMade++;
			numOfMovesLabel.setText(Integer.toString(movesMade));
			
			HanoiCanv.push(toRedo, tmp.to);			
			
			undo.push(tmp);
	//draw disk at previous space then repaint new location occupied by disk
			repaint(prevX, prevY, toRedo.getWidth(), Disk.HEIGHT);			
			repaint(toRedo.x, toRedo.y, toRedo.getWidth(), Disk.HEIGHT);
		
		} else {
			System.out.println("redo invalid this should not happen");	
			redo = StackFactory.createStack();	
		}
		
		enableRedoUndo();
	}
	
	public void doUndo() {
		
		if (undo.size() == 0)
			return;

		UndoRedoEntry tmp = (UndoRedoEntry)undo.pop();
		
		if (tmp.to.top() == tmp.disk) {

			Disk toUndo = tmp.to.pop();
			int prevX = toUndo.x;
			int prevY = toUndo.y;			
	
			movesMade--;
			numOfMovesLabel.setText(Integer.toString(movesMade));
				
			HanoiCanv.push(toUndo, tmp.from);
			
			redo.push(tmp);
	//draw disk at previous space then repaint new location occupied by disk
			repaint(prevX, prevY, toUndo.getWidth(), Disk.HEIGHT);			
			repaint(toUndo.x, toUndo.y, toUndo.getWidth(), Disk.HEIGHT);	
					
		} else {
			System.out.println("undo invalid this is acceptable");
			undo = StackFactory.createStack();	
		}
		
		enableRedoUndo();				
	}	
	
	/**
	 *@return true if mouse is on top of rectangular region false otherwise 
	 */
	 
	private boolean onTopOfRect(int x, int y, int width, int height) {
		// xMo, yMo has global scope
		return (xMo >= x) && (xMo <= x + width) && (yMo >= y) && 
			(yMo <= y + height);
	}
	
	/**
	 *@return true if mouse is on top of param disk
	 */
	 
	private boolean onTopOfDisk(Disk disk) {
		return onTopOfRect(disk.x, disk.y, disk.getWidth(), Disk.HEIGHT);
	}
	
	/**
	 *@return true if mouse is on top of param peg
	 */
	 
	private boolean onTopOfPeg(Peg peg) {
		return onTopOfRect(peg.getX(), peg.getY(), Peg.WIDTH, Peg.HEIGHT);
	}
	
} //##30	
