/*
 * Copyright (c) Developers of JTowers
 * Redistributable and modifiable. Please see license.readme for details.
 *
 * ammendment history
 * <date>			<changed by>		<summary of changes>
 * 18 August 2004	Jonee Ryan Ty		First documented history
 *
 */


import java.awt.Canvas;

import javax.swing.JLabel;


class TimerThread 		extends Thread
{
	public JLabel timerLabel;
	private long seconds; //this time is base 0.01
	public boolean lesserBase = false;	// base can be either 0.01 or 0.1
	private boolean isNotInterrupted = true;
	private int mySleep = 100;

	public TimerThread(JLabel timerLabel, long beginWithSeconds) {
        this.timerLabel = timerLabel;
        this.seconds = beginWithSeconds;
	}
	
	public TimerThread(JLabel timerLabel, long beginWithSeconds, 
			boolean isTenDelay) {
        this.timerLabel = timerLabel;
        this.seconds = beginWithSeconds;
        mySleep = isTenDelay? Constants.FASTEST : 100;
	}
	
	public TimerThread(JLabel timerLabel, long beginWithSeconds, int sleep) {
        this.timerLabel = timerLabel;
        this.seconds = beginWithSeconds;
        mySleep = sleep;        
	}

	public void run() {
		
		while(isNotInterrupted) {
			try {
				this.sleep(mySleep);	// sleep for mySleep / 1000 seconds
			}

			catch(InterruptedException e) {
				isNotInterrupted = false;
			}

			if (lesserBase)
				seconds++;
			else	
				seconds += 10;
			
			try {
				if (!lesserBase)
					timerLabel.setText(Long.toString(seconds / 10));
				else
					timerLabel.setText(Long.toString(seconds));
			}

			catch(NullPointerException npe) {
				isNotInterrupted = false;
			}
		}
	}

}	// end of TimerThread


class TowerThread	extends Thread
{
	public HanoiCanv hanoiCanv;	
	public boolean isInterrupted = false;
	public Peg sandy, dandy, andy;
	
	public TowerThread(HanoiCanv hanoiCanv) {
		
		this.sandy = hanoiCanv.sandy;
		this.andy = hanoiCanv.andy;
		this.dandy = hanoiCanv.dandy;

		this.hanoiCanv = hanoiCanv;	
	}
	
	public void mySleep(long millis) {
		
		try {
			this.sleep(millis);		
		}

		catch(InterruptedException ie) {
			isInterrupted = true;	
		}			
	}
	
	public void moveADisk(Peg pegFrom, Peg pegTo) {
		
		if (isInterrupted)
			return;
		
		Disk moved;
					
		try {
			
			moved = pegFrom.top();
			
			if (hanoiCanv != null) {
				hanoiCanv.aDiskSelected = true;
				hanoiCanv.selected = moved;

				hanoiCanv.repaint(moved.x, moved.y, moved.getWidth(), 
					Disk.HEIGHT);
			
				mySleep(hanoiCanv.delay / 2);
			
			} else
				return;
			
			if (isInterrupted)
				return;
				
			pegFrom.pop();	// no need to store popped because moved is it

			int prevX = moved.x,	// previous coordinates of disk
		 		prevY = moved.y;

			HanoiCanv.push(moved, pegTo);
			moved.whereContained = pegTo;

			if (hanoiCanv != null) {
				hanoiCanv.movesMade++;
				hanoiCanv.numOfMovesLabel.setText(
					Integer.toString(hanoiCanv.movesMade));

				//draw disk at new and previous locations
				hanoiCanv.repaint(moved.x, moved.y, moved.getWidth(), 
					Disk.HEIGHT);
				hanoiCanv.repaint(prevX, prevY, moved.getWidth(), Disk.HEIGHT);
			}
					
		} catch (NullPointerException npe) {
			npe.printStackTrace();
			System.out.println("caught\n");
			return;
		} catch (java.util.EmptyStackException ese) {
			ese.printStackTrace();
			System.out.println("caught\n");			
			return;
		}
		
		if (hanoiCanv != null) {
			mySleep(hanoiCanv.delay / 2);
			if (hanoiCanv != null) {
				hanoiCanv.aDiskSelected = false;
			
				hanoiCanv.repaint(moved.x, moved.y, moved.getWidth(), 
					Disk.HEIGHT);			
			}
		}

		if (isInterrupted)
			return;
	}	
	
} //end of TowerThread


class SolveThread	extends TowerThread
{
	public TimerThread timerThread;

	private JLabel timerLabel;

	public SolveThread(HanoiCanv hanoiCanv) {
		
		super(hanoiCanv);
	}

	public void run() {
		
		hanoiCanv.isSolving = true;
		hanoiCanv.mouseEventsAreBlocked = true;
		
		mySleep(500);	// pause for half a second before starting

		if (isInterrupted) {
			if (hanoiCanv != null) {
				hanoiCanv.isSolving = false;
				hanoiCanv.mouseEventsAreBlocked = true;
			}				
			return;
		}
		
		timerLabel = hanoiCanv.timerLabel;
		
		if (hanoiCanv.delay == Constants.FASTEST)
			timerThread = new TimerThread(timerLabel, 0, true);	
		else
        	timerThread = new TimerThread(timerLabel, 0);

		timerThread.start();

		boolean isConstrained = hanoiCanv.isConstrained;
		boolean isSpread = hanoiCanv.isSpread;
		
		if(!hanoiCanv.isRandomized) {

			if (isConstrained && isSpread)
				hanoiCanv.showMessage(Constants.HELP);
			
			else if (isConstrained)
				constrainedMove(hanoiCanv.numOfDisks, sandy, dandy);
				
			else if (isSpread)
				hanoiCanv.showMessage(Constants.HELP);
				
			else	
				originalMove(hanoiCanv.numOfDisks, sandy, dandy, andy);
		}

		else {				
			
			if (isConstrained && isSpread)
				hanoiCanv.showMessage(Constants.HELP);
			
			else if (isConstrained)
				hanoiCanv.showMessage(Constants.HELP);
			
			else if (isSpread)
				hanoiCanv.showMessage(Constants.HELP);	
			
			else	
				hanoiCanv.showMessage(Constants.HELP);

		}

			// to destroy timerThread if it still exists
		if (timerThread != null) {
			try {
				if (timerThread.isAlive())
					timerThread.interrupt();
			} catch(SecurityException se) {}
				
			timerThread.timerLabel = null;
			timerThread = null;			
		}
		
		if (hanoiCanv != null) {
				// to show ask for help remarks
			hanoiCanv.repaint(0, 0, HanoiCanv.BACK_WIDTH, 50);	
			hanoiCanv.mouseEventsAreBlocked = true;
			hanoiCanv.isSolving = false;	
		}
	}

	private void constrainedMove(int numOfDisks, Peg source, Peg destination)
	{
		if (isInterrupted)
			return;
			
		if(numOfDisks == 0)	// base case
			return;

		if((source == sandy) && (destination == dandy))
		{
			constrainedMove(numOfDisks - 1, sandy, dandy);
			moveADisk(sandy, andy);
			constrainedMove(numOfDisks - 1, dandy, sandy);
			moveADisk(andy, dandy);
			constrainedMove(numOfDisks - 1, sandy, dandy);
		}

		else if((source == dandy) && (destination == sandy))
		{
			constrainedMove(numOfDisks - 1, dandy, andy);
			moveADisk(dandy, sandy);
			constrainedMove(numOfDisks - 1, andy, sandy);
		}

		else if((source == dandy) && (destination == andy))
		{
			constrainedMove(numOfDisks - 1, dandy, sandy);
			moveADisk(dandy, andy);
			constrainedMove(numOfDisks - 1, sandy, andy);
		}

		else if((source == andy) && (destination == sandy))
		{
			constrainedMove(numOfDisks - 1, andy, sandy);
			moveADisk(andy, dandy);
			constrainedMove(numOfDisks - 1, sandy, andy);
			moveADisk(dandy, sandy);
			constrainedMove(numOfDisks - 1, andy, sandy);
		}

		else if((source == sandy) && (destination == andy))
		{
			constrainedMove(numOfDisks - 1, sandy, dandy);
			moveADisk(sandy, andy);
			constrainedMove(numOfDisks - 1, dandy, andy);
		}

		else
		{
			System.out.println("just to make sure everything is accounted for");
		}
	}

	private void originalMove(int numOfDisks, Peg source, Peg destination, 
			Peg alternate) {
				
		if (isInterrupted)
			return;
			
		if(numOfDisks > 0)
		{
			originalMove(numOfDisks - 1, source, alternate, destination);
			moveADisk(source, destination);
			originalMove(numOfDisks - 1, alternate, destination, source);
		}
	}
	
		//changes time to base 0.01 or 0.1 whatever applicable
	public void changeTimeBase(boolean toLesser) {  
		if (timerThread != null)
			timerThread.lesserBase = toLesser;
	}
	
} // end of SolveThread


class SetupThread 	extends TowerThread
{
	private int action = 0;
	public static final int SETUP = 0,
							WILL_SOLVE_NEXT = 1,
							WILL_STEP = 2;
		
	public SetupThread(HanoiCanv hanoiCanv, int action) {
		super(hanoiCanv);	
		this.action = action;
	}	

	public void run() {
		try {		
			if (hanoiCanv != null) {
				hanoiCanv.setup();
			
				Disk tmpDisk;
				Peg whereToPush = null;
				int[] permut, spread;
				int len = 0;
				
				if (hanoiCanv != null) {
					spread = hanoiCanv.spread;
					permut = hanoiCanv.permut;
					len = hanoiCanv.numOfDisks;
				
				} else
					return;
								
				mySleep(300);
				
				if (isInterrupted)
					return;
					
				for (int a = len; a > 0; a--) {
			
					if (hanoiCanv != null) {
						if (spread[a - 1] == HanoiCanv.SANDY)
							whereToPush = hanoiCanv.sandy;
						else if (spread[a - 1] == HanoiCanv.ANDY)
							whereToPush = hanoiCanv.andy;
						else
							whereToPush = hanoiCanv.dandy;
					} else
						return;
					
					tmpDisk = new Disk(
						whereToPush.getMidX() - 
							(Disk.BASE_WIDTH + 
							Disk.ADD_WIDTH_PER_NUM * permut[a - 1]) / 2,
						whereToPush.getBaseY() - 
							(1 + whereToPush.size()) * Disk.HEIGHT, 
						permut[a - 1]);
					
					tmpDisk.whereContained = whereToPush;
					whereToPush.push(tmpDisk);
				
					if (hanoiCanv != null)
						hanoiCanv.repaint(tmpDisk.x, tmpDisk.y, 
							tmpDisk.getWidth(), Disk.HEIGHT);	
				
					mySleep(Constants.ANIM);
					
					if (isInterrupted)
						a = len;
				}
			
			} else
				return;		
									
			if (action == WILL_SOLVE_NEXT) {

				if (hanoiCanv != null) {
					hanoiCanv.solve = new SolveThread(hanoiCanv);
					hanoiCanv.solve.start();
				}
			
			} else if (action == WILL_STEP)
				hanoiCanv.iteration(0);
		
			if (hanoiCanv != null) {
				hanoiCanv.repaint(0, 0, HanoiCanv.BACK_WIDTH, 50);
				hanoiCanv.mouseEventsAreBlocked = false;
			}
			
		} catch(Exception ex) {
			ex.printStackTrace();
			System.out.println("caught\n");
		}
	}

} // SetupThread


